#include <oxstd.h>
#include <oxprob.h>
/**************************************************************************
 * PROGRAM:      PROY.ox  -- ?th revision
 *
 * PURPOSE:
 *               fit of a Multivariate Elliptical Linear Models with known extra parameter 
 *		   
 *
 * REMARKS:      The program uses the BFGS (quasi-Newton) method of 
 *               nonlinear optimization for estimate for location and scale
 *               parameter and estimate for SE basead on Hessian Matrix.
 *
 *                Y_i= X_i\beta+ \epsilon_i  i =1,...,n
 *
 *                \epsilon ~ El(0,\Sigma_i)   , \Sigma_i=Z_i\tau Z_i' +\sigma*In
 *                Y_ is (rx1)-vector
 *                X_i is (rxp)-matrix
 *                Z_i is (px(q-1))-matrix
 *
 * AUTHOR:       Francisco Jos de A. Cysneiros (cysneiros@de.ufpe.br).
 * 				
 *
 * LEGAL NOTE:   This program should NOT be used by or given to anyone 
 *               without explicit permission from the author. Also, the  
 *               author shall not be held liable for any damages that  
 *               possible errors or inaccuracies in the program may cause.
 *
 * VERSION:      1.0 (not ready for prime time yet, but almost...)
 *
 * DATE:         JUNE 09, 2004.
 * 
 * LAST REVISED: .
 */
#include <oxfloat.h>
#import <maximize>
#pragma link("maximize.oxo")
static decl dg;					 // parameter of distribution
static decl fg;					 // parameter of distribution
static decl Y;                   // dependent (response) variable (p x n)
static decl XG;                  // matrix X=(X1',....Xn') matrix of covariates (p x r))
static decl ZG;					 //	matrix Z=(Z1',....Zn')
static decl nuG;				 // degree of freedom
static decl kG;					 // parameter of PE distribution
static decl p;                   // dimension of explanatory variables 
static decl n;		             // number of observations
static decl r;		             // numbers of the location parameters 
static decl q;			         // number of restrictions of scale parameter
static const decl maxit = 10000; // number of iteration maximum			

stder(const vP){
decl Phi,Sigma,vartheta,tau,i,dSigma,m,h,erps,som,B;
dSigma=new array[q];
Sigma=Phi=new array [n];
B=zeros(q-1,q-1);
erps=zeros(q,q);
if (q==1) tau=diag(0); else tau=diag(vP[(r+1):(r+q-1)]);
vartheta=zeros(r,r);
	for(i = 0; i < n; ++i){
	      Sigma[i]= (ZG[i]*tau*ZG[i]')+((vP[r])*unit(p[i]));
		  Phi[i]=invertsym(Sigma[i]);
		  vartheta=vartheta+(4 .*dg[i] ./p[i]) .*(XG[i]' *Phi[i]  * XG[i]);
				}
		  vartheta =  invertsym( vartheta);
		 dSigma[]=zeros(q-1,q-1);
		for(i = 1 ; i<q ; ++i) {
							   B[(i-1)][(i-1)] =1.0;
							   dSigma[i]=B ;
							   }
        erps[0][0]=0.0;
		for(i=0;i<n;++i){
		 som = -0.25*(trace(Phi[i] * unit(int(p[i]))) * trace(	Phi[i] * unit(int(p[i]))))+
		  (fg[i]/(p[i]*(p[i]+2.0)))*(trace(Phi[i] * unit(int(p[i]))) * trace(Phi[i] * unit(int(p[i])))+
		  2* trace(Phi[i] * unit(int(p[i])) *Phi[i]* unit(int(p[i]))));
		  erps[0][0]=erps[0][0]+som;
		  }
	   	for( h = 1 ; h<q ; ++h){
		  erps[0][h]=0.0;
		  for(i=0;i<n;++i){
         som = -0.25*(trace(Phi[i] * unit(int(p[i]))) * trace(	Phi[i] * ZG[i]*dSigma[h]*ZG[i]'))+
		  (fg[i]/(p[i]*(p[i]+2.0)))*(trace(Phi[i] * unit(int(p[i]))) * trace(Phi[i] * ZG[i]*dSigma[h]*ZG[i]')+
		  2* trace(Phi[i] * unit(int(p[i])) * Phi[i] * ZG[i]*dSigma[h]*ZG[i]'));
		 erps[0][h]=erps[0][h]+som;
		 erps[h][0]=erps[0][h];
		 }
		}
	for( m = 1 ; m<q ; ++m){
	   	for( h = 1 ; h<q ; ++h){
		  erps[m][h]=0.0;
		  for(i=0;i<n;++i){
         som = -0.25*(trace(Phi[i] * ZG[i]*dSigma[m]*ZG[i]') * trace(	Phi[i] * ZG[i]*dSigma[h]*ZG[i]'))+
		  (fg[i]/(p[i]*(p[i]+2.0)))*(trace(Phi[i] * ZG[i]*dSigma[m]*ZG[i]') * trace(Phi[i] * ZG[i]*dSigma[h]*ZG[i]')+
		  2* trace(Phi[i] * ZG[i]*dSigma[m]*ZG[i]' * Phi[i] * ZG[i]*dSigma[h]*ZG[i]'));
		 erps[m][h]=erps[m][h]+som;
		 }
		}
	}
	
	erps=invertsym(erps); 
	return(sqrt(diagonal(vartheta))'|sqrt(diagonal(erps))');
	}
	
fLoglikN( const vP, const adFunc, const avScore,const amHessian)
{
decl tau,i,Phi,Sigma,di,mu,dmu,cte,extime;
Sigma=Phi=new array [n];
if (q==1) tau=diag(0); else tau=diag(vP[(r+1):(r+q-1)]);
cte=di = zeros(n,1);

	for(i = 0; i < n; ++i){
	dmu = mu =zeros(int(p[i]),n);
	Sigma[i]= (ZG[i]*tau*ZG[i]')+((vP[r])*unit(p[i]));
    Phi[i]=invert(Sigma[i]);
	mu[][i] = XG[i] *vP[0:(r-1)];
	dmu[][i] = Y[i] - mu[][i];
	di[i][] = (dmu[][i])' * Phi[i] * dmu[][i];
	cte[i][]=  log(determinant(Phi[i]) );
	}
	if((vP[r]<=0) )
	{
	adFunc[0]=-DBL_EPSILON^(-1);
	   return(1);
	 }
      adFunc[0] = double(
    sumc( log(1 ./ (  M_2PI .^ ( p ./ 2.0 ))) + 0.5  * cte -0.5  .* di));//L(sigma,Y)(fdp)	;
	if (isnan(adFunc[0]) || isdotinf(adFunc[0]))
	return 0;
return 1;   // 1 indicates success
}
fLoglikT( const vP, const adFunc, const avScore,const amHessian)
{
decl tau,i,Phi,Sigma,di,mu,dmu,cte,extime;
Sigma=Phi=new array [n];
if (q==1) tau=0; else tau=diag(vP[(r+1):(r+q-1)]);
cte=di = zeros(n,1);
	for(i = 0; i < n; ++i){
	dmu = mu =zeros(int(p[i]),n);
	Sigma[i]= (ZG[i]*tau*ZG[i]')+(vP[r]*unit(p[i]));
    Phi[i]=invert(Sigma[i]);
	mu[][i] = XG[i] *vP[0:(r-1)];
	dmu[][i] = Y[i] - mu[][i];
	di[i][] = (dmu[][i])' * Phi[i] * dmu[][i];
	cte[i]=  log(determinant(Phi[i]) );
	}
	if((vP[r]<=0) )
	{
	adFunc[0]=-DBL_EPSILON^(-1);
	   return(1);
	 }
      adFunc[0] = double(
     sumc( log(gammafact(( nuG + p ) / 2.0) ./ ( ( M_PI .* nuG ) .^ ( p ./ 2.0 ) .*
	  (gammafact( nuG ./ 2.0) ) ) )+ 0.5  *cte) +
	  sumc(-0.5 .* (nuG+p) .* log( 1.0 + di ./ nuG) ) );//L(sigma,Y)(fdp)
	if (isnan(adFunc[0]) || isdotinf(adFunc[0]))
	return 0;
return 1;   // 1 indicates success
}
fLoglikPE( const vP, const adFunc, const avScore,const amHessian)
{
decl tau,i,Phi,Sigma,di,mu,dmu,cte,extime;
Sigma=Phi=new array [n];
if (q==1) tau=0; else tau=diag(vP[(r+1):(r+q-1)]);
cte=di = zeros(n,1);
	for(i = 0; i < n; ++i){
	dmu = mu =zeros(int(p[i]),n);
	Sigma[i]= (ZG[i]*tau*ZG[i]')+(vP[r]*unit(p[i]));
    Phi[i]=invert(Sigma[i]);
	mu[][i] = XG[i] *vP[0:(r-1)];
	dmu[][i] = Y[i] - mu[][i];
	di[i][] = (dmu[][i])' * Phi[i] * dmu[][i];
	cte[i][]=  log(determinant(Phi[i]) );
	}
	if((vP[r]<=0) )
	{
	adFunc[0]=-DBL_EPSILON^(-1);
	   return(1);
	 }
      adFunc[0] = double(
     sumc( log(p)+loggamma(p./2)-0.5*p *log( M_PI) + 0.5  * cte-loggamma(1+(p./(2.0 ./(1+kG))))-
	 (1+(p ./(2.0 ./(1.0+kG))))*log(2) -0.5  .* (di .^(1.0 ./(1.0+kG)))));//L(sigma,Y)(fdp)	;
	if (isnan(adFunc[0]) || isdotinf(adFunc[0]))
	return 0;
return 1;   // 1 indicates success
}
main()
{
decl test,j,mx,m,vP,dfunc,avScore,ir,mhess,extime,dist,err,errI;
	extime = timer();		     // start the timer
 dist=0;                         // if dist=0 Normal
//    dist=1;                    //distribution t-Student
//	  dist=2;                    // Power exponential 
	n=27;
// n=26;	                     // without one obs
 //n=25;	                     //   without two obs
	nuG=zeros(n,1);              //degree of freedom
	kG=zeros(n,1);	             // paramenter k in PE
	Y=new array[n];	             // response 
	p=<4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4>;
	j=0;
	XG=ZG=new array [n];
	   	mx = loadmat("C:/carine/doutorado/ARTIGO_TESE/Exemplos/PROY/programas_corrigidos/PROY.txt");

	for(m=0;m<11;++m){
	Y[m]=mx[j:j+(p[m]-1)][0] ;
	XG[m]=ones(p[m],1)~mx[j:j+(p[m]-1)][1]~zeros(p[m],1)~zeros(p[m],1);
	ZG[m]=ones(p[m],1);	       // one random effect
//	ZG[m]=zeros(p[m],1);	   // independent model
	j=j+p[m];
	nuG[m]=30;
	kG[m]=0.3;
	}
for(m=11;m<n;++m){
	Y[m]=mx[j:j+(p[m]-1)][0] ;
	XG[m]=zeros(p[m],1)~zeros(p[m],1)~ones(p[m],1)~mx[j:j+(p[m]-1)][1];
	ZG[m]=ones(p[m],1);	      // one random effect
//	ZG[m]=zeros(p[m],1);      // independent model
    j=j+p[m];
	nuG[m]=6;
	kG[m]=0.3;
	}

	if(dist==0) {
	          fg=(1.0+0.5 .*p) .*(0.5 .*p);				//for  Normal
	          dg=	p ./4.0;				   			// for Normal
			  }
	else if(dist==1){		  
	fg=(nuG+p) .*(p+2) .*p .*(1 ./(4.0 .*(p+nuG+2)));	// for student-t	
	dg=	(p.*(nuG+p))./(4 .*(nuG+p+2));					//	for student-t
	}
	else {dg=zeros(n,1); fg=zeros(n,1);}

// model matrix					
	r=columns(XG[0]);
	q=columns(ZG[0])+1; 								// component variance model
//  q=1;												// independent model
    errI=zeros(q,1)';  
	mhess=unit(r+q);  									// initialize the matrixHessian/*
	vP=ones(r+q,1); 									//initial value
	MaxControlEps( 1e-6, 1e-4);     					// set the accuracy for BFGS estimation
	    MaxControl(maxit,-1,-1);          				// set maximum # of iterations = 1000
		if(dist==0){
        ir = MaxBFGS(fLoglikN, &vP, &dfunc, &mhess,1);
		if(Num2Derivative(fLoglikN,vP,&mhess)){err=sqrt(diagonal(-invert(mhess)));
	    }
		}
		else
		if(dist==1){ir = MaxBFGS(fLoglikT, &vP, &dfunc,&mhess,1);
		if(Num2Derivative(fLoglikT,vP,&mhess)){err=sqrt(diagonal(-invert(mhess)));}
		}
		else if(dist==2){ir = MaxBFGS(fLoglikPE, &vP, &dfunc, &mhess,1);
		if(Num2Derivative(fLoglikPE,vP,&mhess)){err=sqrt(diagonal(-invert(mhess)));}
		}
		else {println( " ERROR- absent ditribution");}
		if (ir!= MAX_CONV && ir != MAX_WEAK_CONV) test=0; else test=1; //fracass
	print( "\n\t\t Boys (Potthoff) ");
    print( "\n\t\t ------------------\n");
    print( "\n\t\t OX PROGRAM: ", oxfilename(0) );
    print( "\n\t\t OX VERSION: ", oxversion() );
	if(dist==0){
	print( "\n\t\t DISTRIBUTION: Normal ");}
	else 	if(dist==1){
	print( "\n\t\t DISTRIBUTION: T-student D.F Known ");}
    else 	if(dist==2){
	print( "\n\t\t DISTRIBUTION: Power Exponential");}
	else print (" not implemenent yet");
	print( "\n\t\t structured variance : component variance");
	print( "\n\t\t NUM. PARAMETERS IN COV: ", q );
	print( "\n\t\t NUM. PARAMETERS IN MODEL: ", r );
	if (any(p .!= p[0]))  print( "\n\t\t NUM. DIMENSION RESPONSE: unbalanced");
	else print ( "\n\t\t NUM. DIMENSION RESPONSE: balanced p=" ,p[0]);
    print( "\n\t\t NUM. OBSERVATIONS: ", n );
    print( "\n\t\t DATE: ", date() );
    print( "\n\t\t TIME: ", time() );
	print( "\nEXECUTION TIME: ", timespan(extime) , "\n");
		if(test==1) {
	println("\n",MaxConvergenceMsg(ir),"\nFunction value = ", dfunc);
	if(dist!=2) errI=stder(vP)'
		println( "\n PARAMETER ESTIMATES:" );
		print( "%r", {"Estimate param by ML:","EP:","EP(Fisher)"},
	    	   "%c",  {"bo","b1", "b2","b3","sigma","tau's"},
		       "%10.4f", vP'|err |errI);	   
	}
	else{println(" NO CONVERGE");
		println("\n",MaxConvergenceMsg(ir));
	}
	println("---------------------------------------------");
			
}			