/**************************************************************************
 * PROGRAM:      TS-PROY-NO.ox  
 *
 * PURPOSE:		  evaluate the calculation of the TS statistics for the situations 
 *				  mentioned in the paper by assuming Normal distribution for the errors
 *
 *						- independence x random intercept model
 * 						- independence x random slope model
 * 						- independence x random intercept-slope model
 *						- random intercept model x random intercept-slope model
 *
 * AUTHOR:      Carine Savalli (carinesr@ime.usp.br)
 *				Francisco Jos de A. Cysneiros (cysneiros@de.ufpe.br).
 * 				
 *
 * LEGAL NOTE:   This program should NOT be used by or given to anyone 
 *               without explicit permission from the author. Also, the  
 *               author shall not be held liable for any damages that  
 *               possible errors or inaccuracies in the program may cause.
 *
 * VERSION:      1.0 (not ready for prime time yet, but almost...)
 *
 * DATE:         DECEMBER 15, 2004.
 * 
 * LAST REVISED: 
 */


#include <oxstd.h>
#include <oxprob.h>
#include <oxfloat.h>
#import  <solveqp>
main()
{
decl subject,resp,aval,gint,girl,bint,boy,pred;
decl fg,C1,C2,C3,C4,ID,Ji,J,zi,zizi,yi,predi,i,r;
decl Stau1i,Ssigma2i,Ks2s2i,Kt1t1i,Kts2i,Kt1s2i,Stau2i,Kt2t2i; 
decl kt1s2,kts2i,Kt1s2,Stau2,Kt2s2,Kt1t2i,Kt2s2i,sigma,tau1,SIG,INVSIG;
decl Stau1,Ssigma2,Ks2s2,Kt1t1,invKfifi,Kt2t2,Kt1t2,z,TS,TS1;
decl E,G,A,B,C,D;
decl iret,mg,vg,ma,vbs,funr,x;

/*
######## The fitted values of the NORMAL independence model is in predNOef.txt;
*/

decl mx = loadmat("C:/carine/doutorado/artigo_tese/exemplos/proy/programas_corrigidos/predNO.txt");
subject=mx[][0];
resp=mx[][1];
aval=mx[][2];
gint=mx[][3];
girl=mx[][4];
bint=mx[][5];
boy=mx[][6];
pred=mx[][7];
C1=<1.0,0.0,0.0,0.0>;
C2=<0.0,1.0,0.0,0.0>;
C3=<0.0,0.0,1.0,0.0>;
C4=<0.0,0.0,0.0,1.0>;
ID=C1|C2|C3|C4;
Ji=<0.0;0.0;0.0;0.0>;
Ji[0]=1;
Ji[1]=1;
Ji[2]=1;
Ji[3]=1;
J=Ji*Ji';
zi=<0.0;0.0;0.0;0.0>;
zi[0]=8;
zi[1]=10;
zi[2]=12;
zi[3]=14;
zizi=zi*zi';
fg=(4.0*(4.0+2.0))/4.0;


/*
#################################################################################
#######  Calculation of TS statistic for testing one random effect in the########
#######  intercept against independence model                            ########
#################################################################################
*/

yi=zeros(4,1);
predi=zeros(4,1);
SIG=zeros(4,4);

Stau1i=zeros(27,1);
Ssigma2i=zeros(27,1);
Ks2s2i=zeros(27,1);
Kt1t1i=zeros(27,1);
Kt1s2i=zeros(27,1);
sigma=2.214759;
SIG=(sigma^2)*ID;
INVSIG=invert(SIG);

for (i=0;i<27;++i) {

yi[0]=resp[4*i+0];
yi[1]=resp[4*i+1];
yi[2]=resp[4*i+2];
yi[3]=resp[4*i+3];

predi[0]=pred[4*i+0];
predi[1]=pred[4*i+1];
predi[2]=pred[4*i+2];
predi[3]=pred[4*i+3];
r=yi-predi;

Stau1i[i]=(-(1.0/2.0)*sumr(diagonal(INVSIG*J)))  +  ((1.0/2.0)*r'*INVSIG*J*INVSIG*r);

Ssigma2i[i]=(-(1/2)*sumr(diagonal(INVSIG*ID))) +  ((1/2)*r'*INVSIG*ID*INVSIG*r);

Ks2s2i[i]=sumr(diagonal(INVSIG*ID))*sumr(diagonal(INVSIG*ID))*(-1.0/4.0 + fg/(4.0*(4.0+2.0))) + (2.0*fg/(4.0*(4.0+2.0)))*sumr(diagonal(INVSIG*ID*INVSIG*ID));

Kt1t1i[i]=sumr(diagonal(INVSIG*J))*sumr(diagonal(INVSIG*J))*(-1.0/4.0 + fg/(4.0*(4.0+2.0)))   + (2.0*fg/(4.0*(4.0+2.0)))*sumr(diagonal(INVSIG*J*INVSIG*J));

Kt1s2i[i]=sumr(diagonal(INVSIG*J))*sumr(diagonal(INVSIG*ID))*(-1.0/4.0 + fg/(4.0*(4.0+2.0)))  + (2.0*fg/(4.0*(4.0+2.0)))*sumr(diagonal(INVSIG*J*INVSIG*ID));
}
Stau1=sumc(Stau1i);
Ssigma2=sumc(Ssigma2i);
Ks2s2 =sumc(Ks2s2i);
Kt1t1 =sumc(Kt1t1i);
Kt1s2=sumc(Kt1s2i);

invKfifi=1/(Kt1t1 - Kt1s2*(1/Ks2s2)*Kt1s2) ;
 
z=Stau1 - Kt1s2*(1/Ks2s2)*Ssigma2;
//TS=z'*invKfifi*z;
//println("TS independence x random intercept model) = ", TS);

mg = 2*invKfifi;
vg = -2*invKfifi*z;
ma = unit(columns(invKfifi));
vbs = zeros(columns(invKfifi),1);
[iret,x] = SolveQP(mg, vg, ma, vbs, <>, <>, <>, <>);

//funr=(x'mg*x) / 2 + vg'x +z'*invKfifi*z;
//TS1=z'*invKfifi*z-funr;
TS1=-(x'mg*x)/2 - vg'x;
println("TS1 (independence x random intercept model) = ", TS1);

//    SolveQP(const G, const g, const A, const B, const C, const D, const vLo, const vHi); 
//
//    SolveQP solves the quadratic program
//    min f(x)=x'Gx/2 + x'g, subject to:
//    Ax >= b,
//    Cx = d,
//    xlo <= x <= xhi. 




  
/*
#################################################################################
#######  Calculation of TS statistic for testing one random effect in the########
#######  slope against independence model                                ########
#################################################################################
*/

yi=zeros(4,1);
predi=zeros(4,1);
SIG=zeros(4,4);

Stau2i=zeros(27,1);
Ssigma2i=zeros(27,1);
Ks2s2i=zeros(27,1);
Kt2t2i=zeros(27,1);
Kt2s2i=zeros(27,1);
sigma=2.214759;
SIG=(sigma^2)*ID;
INVSIG=invert(SIG);

for (i=0;i<27;++i) {

yi[0]=resp[4*i+0];
yi[1]=resp[4*i+1];
yi[2]=resp[4*i+2];
yi[3]=resp[4*i+3];

predi[0]=pred[4*i+0];
predi[1]=pred[4*i+1];
predi[2]=pred[4*i+2];
predi[3]=pred[4*i+3];
r=yi-predi;
						  
Stau2i[i]= (-(1/2)*sumr(diagonal(INVSIG*zizi)))  +  ((1/2)*r'*INVSIG*zizi*INVSIG*r);

Ssigma2i[i]=(-(1/2)*sumr(diagonal(INVSIG*ID))) +  ((1/2)*r'*INVSIG*ID*INVSIG*r);

Ks2s2i[i]=sumr(diagonal(INVSIG*ID))*sumr(diagonal(INVSIG*ID))*(-1/4 + fg/(4*(4+2))) + (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*ID*INVSIG*ID));

Kt2t2i[i]=sumr(diagonal(INVSIG*zizi))*sumr(diagonal(INVSIG*zizi))*(-1/4 + fg/(4*(4+2)))   + (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*zizi*INVSIG*zizi));

Kt2s2i[i]=sumr(diagonal(INVSIG*zizi))*sumr(diagonal(INVSIG*ID))*(-1/4 + fg/(4*(4+2)))  + (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*zizi*INVSIG*ID));
}
Stau2=sumc(Stau2i);
Ssigma2=sumc(Ssigma2i);
Ks2s2 = sumc(Ks2s2i);
Kt2t2 = sumc(Kt2t2i);
Kt2s2 = sumc(Kt2s2i);
invKfifi=1/(Kt2t2 - Kt2s2*(1/Ks2s2)*Kt2s2);

z=Stau2 - Kt2s2*(1/Ks2s2)*Ssigma2;
//TS=z'*invKfifi*z;
//println("TS  (independence x random slope model)= ", TS);

mg = 2*invKfifi;
vg = -2*invKfifi*z;
ma = unit(columns(invKfifi));
vbs = zeros(columns(invKfifi),1);
[iret,x] = SolveQP(mg, vg, ma, vbs, <>, <>, <>, <>);

//funr=(x'mg*x) / 2 + vg'x +z'*invKfifi*z;
//TS1=z'*invKfifi*z-funr;
TS1=-(x'mg*x)/2 - vg'x;
println("TS1 (independence x random slope model) = ", TS1);


/*
#################################################################################
#######  Calculation of TS statistic for testing two random effect in the########
#######  intercept and/or slope against independence model               ########
#################################################################################
*/

yi=zeros(4,1);
predi=zeros(4,1);
SIG=zeros(4,4);

Stau1i=zeros(27,1);
Stau2i=zeros(27,1);
Ssigma2i=zeros(27,1);
Ks2s2i=zeros(27,1);
Kt1t1i=zeros(27,1);
Kt2t2i=zeros(27,1);
Kt1s2i=zeros(27,1);
Kt2s2i=zeros(27,1);
Kt1t2i=zeros(27,1);
sigma=2.214759;
SIG=(sigma^2)*ID;
INVSIG=invert(SIG);
i=0;

for (i=0;i<27;++i) {

yi[0]=resp[4*i+0];
yi[1]=resp[4*i+1];
yi[2]=resp[4*i+2];
yi[3]=resp[4*i+3];

predi[0]=pred[4*i+0];
predi[1]=pred[4*i+1];
predi[2]=pred[4*i+2];
predi[3]=pred[4*i+3];
r=yi-predi;
   
Stau1i[i]=  (-(1/2)*sumr(diagonal(INVSIG*J)))   +  ((1/2)*r'*INVSIG*J*INVSIG*r);

Stau2i[i]=  (-(1/2)*sumr(diagonal(INVSIG*zizi))) +  ((1/2)*r'*INVSIG*zizi*INVSIG*r);

Ssigma2i[i]=(-(1/2)*sumr(diagonal(INVSIG*ID)))   +  ((1/2)*r'*INVSIG*ID*INVSIG*r);

Ks2s2i[i]=sumr(diagonal(INVSIG*ID))*sumr(diagonal(INVSIG*ID))*(-1/4 + fg/(4*(4+2)))     +  (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*ID*INVSIG*ID)) ;

Kt1t1i[i]=sumr(diagonal(INVSIG*J))*sumr(diagonal(INVSIG*J))*(-1/4 + fg/(4*(4+2)))       +  (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*J*INVSIG*J));

Kt2t2i[i]=sumr(diagonal(INVSIG*zizi))*sumr(diagonal(INVSIG*zizi))*(-1/4 + fg/(4*(4+2))) +  (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*zizi*INVSIG*zizi));

Kt1s2i[i]=sumr(diagonal(INVSIG*J))*sumr(diagonal(INVSIG*ID))*(-1/4 + fg/(4*(4+2)))      +  (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*J*INVSIG*ID));

Kt2s2i[i]=sumr(diagonal(INVSIG*zizi))*sumr(diagonal(INVSIG*ID))*(-1/4 + fg/(4*(4+2)))   +  (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*zizi*INVSIG*ID));

Kt1t2i[i]=sumr(diagonal(INVSIG*J))*sumr(diagonal(INVSIG*zizi))*(-1/4 + fg/(4*(4+2)))    +  (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*J*INVSIG*zizi)) ;
}
Stau1=sumc(Stau1i);
Stau2=sumc(Stau2i) ;
Ssigma2=sumc(Ssigma2i);
Ks2s2 = sumc(Ks2s2i);
Kt1t1 = sumc(Kt1t1i);
Kt2t2 = sumc(Kt2t2i);
Kt1s2=sumc(Kt1s2i);
Kt2s2 = sumc(Kt2s2i);
Kt1t2 = sumc(Kt1t2i);

z=zeros(2,1);
z[0]=Stau1 - Kt1s2*(1/Ks2s2)*Ssigma2;
z[1]=Stau2 - Kt2s2*(1/Ks2s2)*Ssigma2;

invKfifi= zeros(2,2) ;
invKfifi[0][0]=Kt1t1 - Kt1s2*(1/Ks2s2)*Kt1s2;
invKfifi[0][1]=Kt1t2 - Kt1s2*(1/Ks2s2)*Kt2s2;
invKfifi[1][0]=Kt1t2 - Kt2s2*(1/Ks2s2)*Kt1s2;
invKfifi[1][1]=Kt2t2 - Kt2s2*(1/Ks2s2)*Kt2s2;

invKfifi= invert(invKfifi);

//TS=z'*invKfifi*z;
//println("TS  (independence x random intercept-slope model) = ", TS);

mg = 2*invKfifi;
vg = -2*invKfifi*z;
ma = unit(columns(invKfifi));
vbs = zeros(columns(invKfifi),1);
[iret,x] = SolveQP(mg, vg, ma, vbs, <>, <>, <>, <>);

//funr=(x'mg*x) / 2 + vg'x +z'*invKfifi*z;
//TS1=z'*invKfifi*z-funr;
TS1=-(x'mg*x)/2 - vg'x;
println("TS1 (independence x random intercept-slope model) = ", TS1);




/*
#################################################################################
#######  Calculation of TS statistic for testing two random effect in the########
#######  intercept and/or slope against the random intercept model       ########
#################################################################################

######## The fitted values of the NORMAL random-intercept model is in predSTri.txt;
*/

decl mxef = loadmat("C:/carine/doutorado/artigo_tese/exemplos/proy/programas_corrigidos/predNOri.txt");
subject=mxef[][0];
resp=mxef[][1];
aval=mxef[][2];
gint=mxef[][3];
girl=mxef[][4];
bint=mxef[][5];
boy=mxef[][6];
pred=mxef[][7];

yi=zeros(4,1);
predi=zeros(4,1);
SIG=zeros(4,4);

Stau1i=zeros(27,1);
Stau2i=zeros(27,1);
Ssigma2i=zeros(27,1);
Ks2s2i=zeros(27,1);
Kt1t1i=zeros(27,1);
Kt2t2i=zeros(27,1);
Kt1s2i=zeros(27,1);
Kt2s2i=zeros(27,1);
Kt1t2i=zeros(27,1);
sigma=1.369159;
tau1=1.740851;
SIG=sigma^2*ID + tau1^2*J;
INVSIG=invert(SIG);


for (i=0;i<27;++i) {

yi[0]=resp[4*i+0];
yi[1]=resp[4*i+1];
yi[2]=resp[4*i+2];
yi[3]=resp[4*i+3];

predi[0]=pred[4*i+0];
predi[1]=pred[4*i+1];
predi[2]=pred[4*i+2];
predi[3]=pred[4*i+3];
r=yi-predi;

Stau1i[i]=  (-(1/2)*sumr(diagonal(INVSIG*J)))   +  ((1/2)*r'*INVSIG*J*INVSIG*r);

Stau2i[i]=  (-(1/2)*sumr(diagonal(INVSIG*zizi))) +  ((1/2)*r'*INVSIG*zizi*INVSIG*r);

Ssigma2i[i]=(-(1/2)*sumr(diagonal(INVSIG*ID)))   +  ((1/2)*r'*INVSIG*ID*INVSIG*r);


Ks2s2i[i]=sumr(diagonal(INVSIG*ID))*sumr(diagonal(INVSIG*ID))*(-1/4 + fg/(4*(4+2)))    +  (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*ID*INVSIG*ID));

Kt1t1i[i]=sumr(diagonal(INVSIG*J))*sumr(diagonal(INVSIG*J))*(-1/4 + fg/(4*(4+2)))      +  (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*J*INVSIG*J)) ;

Kt2t2i[i]=sumr(diagonal(INVSIG*zizi))*sumr(diagonal(INVSIG*zizi))*(-1/4 + fg/(4*(4+2))) +  (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*zizi*INVSIG*zizi)) ;

Kt1s2i[i]=sumr(diagonal(INVSIG*J))*sumr(diagonal(INVSIG*ID))*(-1/4 + fg/(4*(4+2)))     +  (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*J*INVSIG*ID))	;

Kt2s2i[i]=sumr(diagonal(INVSIG*zizi))*sumr(diagonal(INVSIG*ID))*(-1/4 + fg/(4*(4+2)))  +  (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*zizi*INVSIG*ID));

Kt1t2i[i]=sumr(diagonal(INVSIG*J))*sumr(diagonal(INVSIG*zizi))*(-1/4 + fg/(4*(4+2)))   +  (2*fg/(4*(4+2)))*sumr(diagonal(INVSIG*J*INVSIG*zizi));

}

Stau1=sumc(Stau1i) ;
Stau2=sumc(Stau2i);
Ssigma2=sumc(Ssigma2i);
Ks2s2 = sumc(Ks2s2i);
Kt1t1 = sumc(Kt1t1i);
Kt2t2 = sumc(Kt2t2i);
Kt1s2=sumc(Kt1s2i);
Kt2s2 = sumc(Kt2s2i);
Kt1t2 = sumc(Kt1t2i);

E=Kt1t1 - Kt1s2*invert(Ks2s2)*Kt1s2	;
G=invert(Ks2s2)*Kt1s2;

A=invert(Ks2s2) - G*invert(E)*G';
B=-G*invert(E);
C=B';
D=invert(E);

z=(Stau2 - (Kt2s2*A + Kt1t2*C)*Ssigma2 - (Kt2s2*B + Kt1t2*D)*Stau1)	;

invKfifi=1/( Kt2t2 - Kt2s2*(Kt2s2*A + Kt1t2*C) - Kt1t2*(Kt2s2*B + Kt1t2*D))	;

//TS=z'*invKfifi*z ;
//println("TS =  (random intercept model x random intercept-slope model)", TS);

mg = 2*invKfifi;
vg = -2*invKfifi*z;
ma = unit(columns(invKfifi));
vbs = zeros(columns(invKfifi),1);
[iret,x] = SolveQP(mg, vg, ma, vbs, <>, <>, <>, <>);

//funr=(x'mg*x) / 2 + vg'x +z'*invKfifi*z;
//TS1=z'*invKfifi*z-funr;
TS1=-(x'mg*x)/2 - vg'x;
println("TS1 (random intercept model x random intercept-slope model) = ", TS1);



}
